-- shuffle64.vhdl -- 64-Bit F-CPU Bit Shuffling Unit
-- Copyright (C) 2001 Michael Riepe <michael@stud.uni-hannover.de>
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-- $Id: shuffle64.vhdl,v 1.4 2001/09/22 18:31:47 michael Exp $

library IEEE;
use IEEE.std_logic_1164.all;
use work.Bit_Manipulation.all;

entity Shuffle64 is
	generic (
		WIDTH : natural := 64
	);
	port (
		-- shiftee
		A : in std_ulogic_vector(WIDTH-1 downto 0);
		-- shift count
		B : in std_ulogic_vector(WIDTH-1 downto 0);
		-- operating mode (mutually exclusive)
		ShiftL : in std_ulogic;
		ShiftR : in std_ulogic;
		ShiftRA : in std_ulogic;
		RotL : in std_ulogic;
		RotR : in std_ulogic;
		Bitrev : in std_ulogic;
		Byterev : in std_ulogic;
		Sdup : in std_ulogic;
		Mix : in std_ulogic;
		Expand : in std_ulogic;
		-- `high' flag for mix/expand instructions
		High : in std_ulogic;
		-- SIMD mode flags
		U : in std_ulogic_vector(2 downto 0);
		-- clock/reset/enable inputs (unused)
		Clk : in std_ulogic;
		Rst : in std_ulogic;
		En : in std_ulogic;
	--
		-- output
		Y : out std_ulogic_vector(WIDTH-1 downto 0)
	);
--pragma synthesis_off
begin
	assert WIDTH = 64
		report "WIDTH must be 64"
		severity failure;
--pragma synthesis_on
end Shuffle64;

architecture Behave_1 of Shuffle64 is
	-- byte-wide 16:1 mux
	function ab_sel (aa : in std_ulogic_vector(WIDTH-1 downto 0);
					 bb : in std_ulogic_vector(WIDTH-1 downto 0);
					 se : in std_ulogic_vector(3 downto 0)) return std_ulogic_vector is
		variable yy : std_ulogic_vector(7 downto 0);
	begin
		case se is
			when X"0" => yy := aa( 7 downto  0);
			when X"1" => yy := aa(15 downto  8);
			when X"2" => yy := aa(23 downto 16);
			when X"3" => yy := aa(31 downto 24);
			when X"4" => yy := aa(39 downto 32);
			when X"5" => yy := aa(47 downto 40);
			when X"6" => yy := aa(55 downto 48);
			when X"7" => yy := aa(63 downto 56);
			when X"8" => yy := bb( 7 downto  0);
			when X"9" => yy := bb(15 downto  8);
			when X"A" => yy := bb(23 downto 16);
			when X"B" => yy := bb(31 downto 24);
			when X"C" => yy := bb(39 downto 32);
			when X"D" => yy := bb(47 downto 40);
			when X"E" => yy := bb(55 downto 48);
			when X"F" => yy := bb(63 downto 56);
			when others => yy := (others => 'X'); -- don't care
		end case;
		return yy;
	end ab_sel;

	-- byte-wide 12:1 mux (for upper part of bit-shifted data)
	function hi_sel (aa : in std_ulogic_vector(WIDTH-1 downto 0);
					 xx : in std_ulogic_vector(WIDTH/8-1 downto 0);
					 se : in std_ulogic_vector(3 downto 0)) return std_ulogic_vector is
		variable yy : std_ulogic_vector(7 downto 0);
	begin
		case se is
			when X"0" => yy := aa( 7 downto  0);
			when X"1" => yy := aa(15 downto  8);
			when X"2" => yy := aa(23 downto 16);
			when X"3" => yy := aa(31 downto 24);
			when X"4" => yy := aa(39 downto 32);
			when X"5" => yy := aa(47 downto 40);
			when X"6" => yy := aa(55 downto 48);
			when X"7" => yy := aa(63 downto 56);
			when X"9" => yy := (others => xx(1));
			when X"B" => yy := (others => xx(3));
			when X"D" => yy := (others => xx(5));
			when X"F" => yy := (others => xx(7));
			when others => yy := (others => 'X'); -- don't care
		end case;
		return yy;
	end hi_sel;

	-- byte-wide 9:1 mux (for lower part of bit-shifted data)
	function lo_sel (aa : in std_ulogic_vector(WIDTH-1 downto 0);
					 se : in std_ulogic_vector(3 downto 0)) return std_ulogic_vector is
		variable yy : std_ulogic_vector(7 downto 0);
	begin
		case se is
			when X"0" => yy := aa( 7 downto  0);
			when X"1" => yy := aa(15 downto  8);
			when X"2" => yy := aa(23 downto 16);
			when X"3" => yy := aa(31 downto 24);
			when X"4" => yy := aa(39 downto 32);
			when X"5" => yy := aa(47 downto 40);
			when X"6" => yy := aa(55 downto 48);
			when X"7" => yy := aa(63 downto 56);
			when X"8" => yy := (others => '0');
			when others => yy := (others => 'X'); -- don't care
		end case;
		return yy;
	end lo_sel;

	-- basic bitwise right shift/rotate operation
	-- NOTE: all bitwise operations use this.
	function simd_shuffle (A, B, U : in std_ulogic_vector;
						   Rot, Ext : in std_ulogic) return std_ulogic_vector is
		constant w : natural := A'length;
		alias aa : std_ulogic_vector(w-1 downto 0) is A;
		alias bb : std_ulogic_vector(B'length-1 downto 0) is B;
		alias uu : std_ulogic_vector(U'length-1 downto 0) is U;
		variable ah : std_ulogic_vector(w-1 downto 0);
		variable al : std_ulogic_vector(w-1 downto 0);
		variable yy : std_ulogic_vector(w-1 downto 0);
		variable bsc : std_ulogic_vector(6 downto 0);
		variable tt : std_ulogic_vector(15 downto 0);
		variable ee : std_ulogic_vector(w/8-1 downto 0);
		variable k : natural;
		variable hi, lo : std_ulogic_vector(4*(w/8)-1 downto 0);
		variable sel : std_ulogic_vector(2 downto 0);
	begin
--pragma synthesis_off
		--assert w mod 8 = 0 report "w must be a multiple of 8" severity failure;
		assert w = 64 report "w must be 64" severity failure;
		assert B'length >= 6 report "B is too short" severity failure;
		assert U'length >= 3 report "U is too short" severity failure;
--pragma synthesis_on

		-- SIMD sign extension
		ee := (others => '1');
		for j in uu'range loop
			k := 2 ** j;
			for i in ee'length/k/2-1 downto 0 loop
				if 2*i*k+k-1 < ee'length then
					ee(2*i*k+k-1) := not uu(j);
				end if;
			end loop;
		end loop;
		-- XXX: speed this up?
		for i in w/8-1 downto 0 loop
			ee(i) := ee(i) and Ext and not Rot and aa(8*i+7);
		end loop;
		ee := to_X01(ee);

		-- bit-shift individual bytes
		sel := bb(2 downto 0);
		sel := to_X01(sel);
		for i in w/8-1 downto 0 loop
			tt := (others => '0');
			tt(15 downto 8) := aa(8*i+7 downto 8*i);
			tt := to_X01(tt);
			case sel is
				when "000" => tt := rshift(tt, 0, ee(i));
				when "001" => tt := rshift(tt, 1, ee(i));
				when "010" => tt := rshift(tt, 2, ee(i));
				when "011" => tt := rshift(tt, 3, ee(i));
				when "100" => tt := rshift(tt, 4, ee(i));
				when "101" => tt := rshift(tt, 5, ee(i));
				when "110" => tt := rshift(tt, 6, ee(i));
				when "111" => tt := rshift(tt, 7, ee(i));
				when others => tt := (others => 'X'); -- don't care
			end case;
			ah(8*i+7 downto 8*i) := tt(15 downto 8);
			al(8*i+7 downto 8*i) := tt( 7 downto 0);
		end loop;

		-- calculate byte selector
		bsc(6) := Rot;
		bsc(5 downto 3) := uu(2 downto 0);
		bsc(2 downto 0) := bb(5 downto 3) and uu(2 downto 0);
		bsc := to_X01(bsc);
		case bsc is
			-- shift.8
			when "0000000" => hi := X"76543210"; lo := X"88888888";
			-- rot.8
			when "1000000" => hi := X"76543210"; lo := X"76543210";
			-- shift.16
			when "0001000" => hi := X"76543210"; lo := X"87858381";
			when "0001001" => hi := X"F7D5B391"; lo := X"88888888";
			-- rot.16
			when "1001000" => hi := X"76543210"; lo := X"67452301";
			when "1001001" => hi := X"67452301"; lo := X"76543210";
			-- shift.32
			when "0011000" => hi := X"76543210"; lo := X"87658321";
			when "0011001" => hi := X"F765B321"; lo := X"88768832";
			when "0011010" => hi := X"FF76BB32"; lo := X"88878883";
			when "0011011" => hi := X"FFF7BBB3"; lo := X"88888888";
			-- rot.32
			when "1011000" => hi := X"76543210"; lo := X"47650321";
			when "1011001" => hi := X"47650321"; lo := X"54761032";
			when "1011010" => hi := X"54761032"; lo := X"65472103";
			when "1011011" => hi := X"65472103"; lo := X"76543210";
			-- shift.64
			when "0111000" => hi := X"76543210"; lo := X"87654321";
			when "0111001" => hi := X"F7654321"; lo := X"88765432";
			when "0111010" => hi := X"FF765432"; lo := X"88876543";
			when "0111011" => hi := X"FFF76543"; lo := X"88887654";
			when "0111100" => hi := X"FFFF7654"; lo := X"88888765";
			when "0111101" => hi := X"FFFFF765"; lo := X"88888876";
			when "0111110" => hi := X"FFFFFF76"; lo := X"88888887";
			when "0111111" => hi := X"FFFFFFF7"; lo := X"88888888";
			-- rot.64
			when "1111000" => hi := X"76543210"; lo := X"07654321";
			when "1111001" => hi := X"07654321"; lo := X"10765432";
			when "1111010" => hi := X"10765432"; lo := X"21076543";
			when "1111011" => hi := X"21076543"; lo := X"32107654";
			when "1111100" => hi := X"32107654"; lo := X"43210765";
			when "1111101" => hi := X"43210765"; lo := X"54321076";
			when "1111110" => hi := X"54321076"; lo := X"65432107";
			when "1111111" => hi := X"65432107"; lo := X"76543210";
			-- don't care
			when others   => hi := (others => 'X'); lo := (others => 'X');
		end case;

		-- output stage
		for i in w/8-1 downto 0 loop
			yy(8*i+7 downto 8*i) := hi_sel(ah, ee, hi(4*i+3 downto 4*i))
								 or lo_sel(al, lo(4*i+3 downto 4*i));
		end loop;
		return yy;
	end simd_shuffle;

	subtype mode_type is std_ulogic_vector(10 downto 0);

	constant MODE_SHIFTL  : mode_type := "10000000000";
	constant MODE_SHIFTR  : mode_type := "01000000000";
	constant MODE_SHIFTRA : mode_type := "00100000000";
	constant MODE_ROTL    : mode_type := "00010000000";
	constant MODE_ROTR    : mode_type := "00001000000";
	constant MODE_BITREV  : mode_type := "00000100000";
	constant MODE_BYTEREV : mode_type := "00000010000";
	constant MODE_SDUP    : mode_type := "00000001000";
	constant MODE_MIXL    : mode_type := "00000000100";
	constant MODE_MIXH    : mode_type := "00000000101";
	constant MODE_EXPANDL : mode_type := "00000000010";
	constant MODE_EXPANDH : mode_type := "00000000011";

	signal Mode : mode_type;

	signal Y_ShiftRotL : std_ulogic_vector(WIDTH-1 downto 0);
	signal Y_ShiftRotR : std_ulogic_vector(WIDTH-1 downto 0);
	signal Y_Bitrev : std_ulogic_vector(WIDTH-1 downto 0);
	signal Y_Bytewise : std_ulogic_vector(WIDTH-1 downto 0);
begin
	-- mode vector
	Mode <= (
		10 => ShiftL,
		 9 => ShiftR,
		 8 => ShiftRA,
		 7 => RotL,
		 6 => RotR,
		 5 => Bitrev,
		 4 => Byterev,
		 3 => Sdup,
		 2 => Mix,
		 1 => Expand,
		 0 => High,
		others => 'X'
	);

	-- shift / rotate / bitrev
	process (A, B, U, ShiftRA, RotL, RotR)
		variable xx : std_ulogic_vector(WIDTH-1 downto 0);
		variable yy : std_ulogic_vector(WIDTH-1 downto 0);
		variable uu : std_ulogic_vector(2 downto 0);
	begin
		-- shiftr / shiftra / rotr
		xx := simd_shuffle(A, B, U, RotR, ShiftRA);
		Y_ShiftRotR <= xx;
		-- shiftl / rotl
		-- shiftla ("arithmetic" left-shift) is also possible
		xx := bit_reverse(A);
		xx := simd_shuffle(xx, B, U, RotL, '0');
		xx := bit_reverse(xx);
		Y_ShiftRotL <= xx;
		-- bitrev
		-- NOTE: `rev(shl(x))' = `shr(rev(x))'
		yy := (others => '0');
		uu := to_X01(U);
		case uu is
			when "000" => yy( 7 downto 0) := bit_reverse(xx( 7 downto 0));
			when "001" => yy(15 downto 0) := bit_reverse(xx(15 downto 0));
			when "011" => yy(31 downto 0) := bit_reverse(xx(31 downto 0));
			when "111" => yy(63 downto 0) := bit_reverse(xx(63 downto 0));
			when others => yy := (others => 'X');
		end case;
		Y_Bitrev <= yy;
	end process;

	-- bytewise stuff (byterev, sdup, mix, expand)
	process (A, B, U, Byterev, Sdup, Mix, Expand, High)
		variable sel : std_ulogic_vector(7 downto 0);
		variable xx : std_ulogic_vector(WIDTH/2-1 downto 0);
		variable yy : std_ulogic_vector(WIDTH-1 downto 0);
	begin
		sel(7) := Byterev;
		sel(6) := Sdup;
		sel(5) := Mix;
		sel(4) := Expand;
		sel(3) := High;
		sel(2 downto 0) := U;
		sel := to_X01(sel);
		case sel is
			-- byterev
			when "10000000" => xx := X"76543210";
			when "10000001" => xx := X"67452301";
			when "10000011" => xx := X"45670123";
			when "10000111" => xx := X"01234567";
			-- sdup
			when "01000000" => xx := X"00000000";
			when "01000001" => xx := X"10101010";
			when "01000011" => xx := X"32103210";
			when "01000111" => xx := X"76543210";
			-- mixh
			-- XXX: swap input registers?
			when "00101000" => xx := X"F7E6D5C4";
			when "00101001" => xx := X"FE76DC54";
			when "00101011" => xx := X"FEDC7654";
			when "00101111" => xx := X"FEDCBA98"; -- XXX: undefined
			-- mixl
			-- XXX: swap input registers?
			when "00100000" => xx := X"B3A29180";
			when "00100001" => xx := X"BA329810";
			when "00100011" => xx := X"BA983210";
			when "00100111" => xx := X"76543210"; -- XXX: undefined
			-- expandh
			-- XXX: swap input registers?
			when "00011000" => xx := X"F7D5B391";
			when "00011001" => xx := X"FE76BA32";
			when "00011011" => xx := X"FEDC7654";
			when "00011111" => xx := X"FEDCBA98"; -- XXX: undefined
			-- expandl
			-- XXX: swap input registers?
			when "00010000" => xx := X"E6C4A280";
			when "00010001" => xx := X"DC549810";
			when "00010011" => xx := X"BA983210";
			when "00010111" => xx := X"76543210"; -- XXX: undefined
			-- don't care
			when others => xx := (others => 'X');
		end case;
		for i in WIDTH/8-1 downto 0 loop
			yy(8*i+7 downto 8*i) := ab_sel(A, B, xx(4*i+3 downto 4*i));
		end loop;
		Y_Bytewise <= yy;
	end process;

	-- output mux
	process (Mode, Y_ShiftRotL, Y_ShiftRotR, Y_Bitrev, Y_Bytewise)
		variable mm : mode_type;
		variable yy : std_ulogic_vector(WIDTH-1 downto 0);
	begin
		mm := to_X01(Mode);
		case mm is
			when MODE_SHIFTL | MODE_ROTL =>
				yy := Y_ShiftRotL;
			when MODE_SHIFTR | MODE_SHIFTRA | MODE_ROTR =>
				yy := Y_ShiftRotR;
			when MODE_BITREV =>
				yy := Y_Bitrev;
			when MODE_BYTEREV | MODE_SDUP | MODE_MIXL
			   | MODE_MIXH | MODE_EXPANDL | MODE_EXPANDH =>
				yy := Y_Bytewise;
			when others =>
				yy := (others => 'X'); -- don't care
		end case;
		Y <= yy;
	end process;
end Behave_1;

-- vi: set ts=4 sw=4 equalprg="fmt -72 -p--": please
